# %%
from typing import *
from toolz.curried import pipe, take, map, merge, curry, compose_left
import itertools as it
from functools import singledispatch
import pandas as pd
import re

__all__ = "to_table transpose to_date to_num to_camel decamel add_under".split()

# %%
@singledispatch
def to_table(data: Iterable[Dict[str, Any]]) -> Iterator[Dict[str, Any]]:
    """Conversion function that takes either a
        pandas DataFrame,
        Iterable of Dicts,
        dict,
        or string
    and converts it into an iterator (generator) of records where each record
    is a dict with keys of type str and values of type Any.

    >>> data = [
    ...     {"fname": "Axl", "lname": "Rose", "Band": "Guns 'n' Roses"},
    ...     {"fname": "Angus", "lname": "Young", "Band": "AC/DC"},
    ...     {"fname": "Steven", "lname": "Tyler", "Band": "Aerosmith"},
    ... ]
    # Returns an iterator (generator)
    >>> to_table(data)
    <generator object to_table at 0x10c101eb0>
    # Use list to visualize the data
    >>> list(to_table(data))
    [
        {'fname': 'Axl', 'lname': 'Rose', 'Band': "Guns 'n' Roses"},
        {'fname': 'Angus', 'lname': 'Young', 'Band': 'AC/DC'},
        {'fname': 'Steven', 'lname': 'Tyler', 'Band': 'Aerosmith'}
    ]
    >>> list(to_table(data[0]))
    [
        {'column': 'fname', 'value': 'Axl'},
        {'column': 'lname', 'value': 'Rose'},
        {'column': 'Band', 'value': "Guns 'n' Roses"}
    ]
    >>> list(to_table("Axl Rose"))
    [{'value': 'Axl Rose'}]
    >>> df = pd.DataFrame.from_records(data)
    >>> df
        fname  lname            Band
    0     Axl   Rose  Guns 'n' Roses
    1   Angus  Young           AC/DC
    2  Steven  Tyler       Aerosmith
    >>> list(to_table(df))
    [
        {'fname': 'Axl', 'lname': 'Rose', 'Band': "Guns 'n' Roses"},
        {'fname': 'Angus', 'lname': 'Young', 'Band': 'AC/DC'},
        {'fname': 'Steven', 'lname': 'Tyler', 'Band': 'Aerosmith'}
    ]
    """
    for row in data:
        yield row


@to_table.register(dict)
def _(data: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
    for k, v in data.items():
        yield {"column": k, "value": v}


@to_table.register(str)
def _(data: str) -> Iterator[Dict[str, Any]]:
    yield {"value": data}


@to_table.register(pd.DataFrame)
def _(data: pd.DataFrame) -> Iterator[Dict[str, Any]]:
    for row in data.to_dict("records"):
        yield row


# %%
@singledispatch
def transpose(data: Iterable[Dict[str, Any]], limit: int=3
    ) -> Iterable[Dict[str,Any]]:
    """Transposes a table (list of dicts) such that columns are
    rows and rows are columns.

    from atek.conversion import *
    >>> import pandas as pd
    >>> data = [
    ...     {"fname": "Axl", "lname": "Rose", "Band": "Guns 'n' Roses"},
    ...     {"fname": "Angus", "lname": "Young", "Band": "AC/DC"},
    ...     {"fname": "Steven", "lname": "Tyler", "Band": "Aerosmith"},
    ... ]
    >>> list(transpose(data))
    [
        {'column': 'fname', 'row 1': 'Axl', 'row 2': 'Angus', 'row 3': 'Steven'},
        {'column': 'lname', 'row 1': 'Rose', 'row 2': 'Young', 'row 3': 'Tyler'},
        {'column': 'Band', 'row 1': "Guns 'n' Roses", 'row 2': 'AC/DC', 'row 3': 'Aerosmith'}
    ]
    >>> list(transpose(data[0]))
    [
        {'column': 'fname', 'row 1': 'Axl'},
        {'column': 'lname', 'row 1': 'Rose'},
        {'column': 'Band', 'row 1': "Guns 'n' Roses"}
    ]
    >>> list(transpose("Axl Rose"))
    [{'column': "'Axl Rose'", 'row 1': 'Axl Rose'}]
    >>> pd.DataFrame.from_records(data)
        fname  lname            Band
    0     Axl   Rose  Guns 'n' Roses
    1   Angus  Young           AC/DC
    2  Steven  Tyler       Aerosmith
    >>> df = pd.DataFrame.from_records(data)
    >>> df
        fname  lname            Band
    0     Axl   Rose  Guns 'n' Roses
    1   Angus  Young           AC/DC
    2  Steven  Tyler       Aerosmith
    >>> transpose(df)
    column           row 0  row 1      row 2
    0  fname             Axl  Angus     Steven
    1  lname            Rose  Young      Tyler
    2   Band  Guns 'n' Roses  AC/DC  Aerosmith
    """

    # Add a row number to each row of the data
    count = it.count(1)
    row_num = lambda: next(count)

    # Put the keys into a list
    header = lambda d: list(d.keys())

    # Put the list of values from 1 row into a list of many rows where each
    # value is a single row
    values = lambda d: list(zip(*d.values()))

    # Add the header value to each set of rows
    combine = lambda d: [dict(zip(header(d), row)) for row in values(d)]

    # Return the Table as a list of dict records
    return pipe(
        data
        ,take(limit)
        ,map(lambda row: list(zip(*row.items())))
        ,map(lambda row: dict(zip(["column", f"row {row_num()}"], row)))

        # Merge each record into 1 dict where the keys are the column
        # and row numbers
        ,merge
        ,combine
    )


@transpose.register(pd.DataFrame)
def _(data: pd.DataFrame, limit: int=3) -> pd.DataFrame:
    return pipe(
        data
        .head(limit)

        # Switch columns to rows and rows to columns
        .transpose()

        # Reset the index so that the column containing for column names
        # is named 'index'
        .reset_index()

        # Rename columns using index number as the 'row {col}' part
        ,lambda df: df.rename(columns={
            col: ("column" if col == "index" else f"row {int(col)+1}")
            for col in df.columns
        })
    )


@transpose.register(dict)
def _(data: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
    for k, v in data.items():
        yield {"column": k, "row 1": v}


@transpose.register(str)
def _(data: str) -> Iterator[Dict[str, Any]]:
    yield {"value": data, "row 1": str(data)}

transpose = curry(transpose)

# %%
@curry
def to_date(pat: str, df: pd.DataFrame) -> pd.DataFrame:
    """Converts a 1 or more series in a DataFrame to a datetime64
    by applying pandas.to_datetime(errors_"coerce").

    pat arg will match column names. 'date' will convert any column in the
    dataframe with the word 'date' to a datetime64 column.
    """
    return df.assign(**{
        col: pd.to_datetime(df[col], errors="coerce")
        for col in df.columns
        if re.search(pat, col)
    })

# %%
@curry
def to_num(pat: str, df: pd.DataFrame) -> pd.DataFrame:
    """Converts a 1 or more series in a DataFrame to a float64
    by applying pandas.to_numeric(errors_"coerce").

    pat arg will match column names. '*fee*' will convert any column in the
    dataframe with the word 'fee' to a float64 column.
    """
    return df.assign(**{
        col: pd.to_numeric(df[col], errors="coerce")
        for col in df.columns
        if re.search(pat, col)
    })


# %%
@curry
def to_camel(df: pd.DataFrame) -> pd.DataFrame:
    """Converts columns in a dataframe to camel case.

    If df has columns OrderDate, LoanNumber, TrackingNumber
    >>>list(df.columns)
    ["OrderDate", "LoanNumber", "TrackingNumber"]
    >>>to_camel(df)
    ["order_date", "loan_number", "tracking_number"]
    """
    resub = curry(re.sub)

    fix = compose_left(
        resub("\W","")
        # ,resub("_", "")
        ,resub("(?P<digit1>^\d)", "x\g<digit1>")
        ,resub("(?P<lower>[a-z])(?P<upper>[A-Z])", "\g<lower>_\g<upper>")
        ,str.lower
    )
    return df.rename(columns={col: fix(col) for col in df.columns})


# %%
@curry
def decamel(df: pd.DataFrame, *, pats: Optional[Union[List[str],str]]=None
    ) -> pd.DataFrame:
    """Takes columns in camel case and converts them to capitlized versions with
    the underscore replaced with a space and 'pct' replaced with '%' and
    pats replaced with upper case versions.

    >>>list(df.columns)
    ["order_date", "tracking_id", "pct_orders"]
    >>>list(decamel(["id"], df).columns)
    ["Order Date", "Tracking ID", "% Orders"]
    """
    pats = (
        [pats] if isinstance(pats, str) else
        [] if pats is None else
        pats
    )
    return df.rename(columns={
        col: " ".join([
            (part.upper() if part in pats else part.capitalize())
            .replace("Pct", "%")
            .replace("#", "No.")
            for part in col.split("_")
        ])
        for col in df.columns
    })


# %%
@curry
def add_under(pat: str, df: pd.DataFrame) -> pd.DataFrame:
    """Addes an underscore character before 'pat' if 'pat' is found at the end
    of the column name.

    >>>list(df.columns)
    ["trackingid", "loan_number", "idtracking"]
    >>>list(dunder(df).columns)
    ["tracking_id", "loan_number", "idtracking"]
    """
    return df.rename(columns={
        col: re.sub(
            f"(?P<lower>[a-z])(?P<pat>{pat})$",
            "\g<lower>_\g<pat>",
            col
        ).lower()
        for col in df.columns
    })




# %% imports
import pandas as pd
import toolz.curried as tz
from pathlib import Path
import atek
from typing import Union, Dict, Optional
from dataclasses import dataclass, asdict
from fnmatch import fnmatch
import xlsxwriter.utility as xlcell

__all__ = ["xlformat", "to_xlsx", "XLCell", "XLRange", "data_range"]

# %% calc_column_width
def calc_column_width(data: pd.DataFrame, min_width: int=10, max_width: int=50
    ) -> Dict[str,int]:
    """Uses converts values to strings and gets the max length of all values
    and column names by column."""
    value_width = {
        col: data[col]
            .apply(lambda x: len(str(x)))
            .max()
        for col in data.columns
    }
    header_width = {col: len(col) for col in data.columns}
    min_width = {col: min_width for col in data.columns}
    max_width = {col: max_width for col in data.columns}
    width = tz.pipe(
        tz.merge_with(
            max
            ,value_width
            ,header_width
            ,min_width
        )
        ,lambda d: tz.merge_with(min, d, max_width)
    )
    return width

# calc_column_width(data, 4, 20)


# %% Codes
Codes = Dict[str, Union[str,int]]
# %% XLFormat
@dataclass(frozen=True)
class XLFormat:
    pattern: str
    codes: Codes


# %% ColFormat
@dataclass(frozen=True)
class ColFormat:
    header_codes: Codes
    value_codes: Codes


# %% XLPos
@dataclass(frozen=True)
class XLCell:
    row: int=0
    col: int=0
    row_abs: bool=False
    col_abs: bool=False

    @classmethod
    def from_cell(cls, cell: str):
        row, col, row_abs, col_abs = xlcell.xl_cell_to_rowcol_abs(cell)
        return cls(row, col, row_abs, col_abs)

    @classmethod
    def from_rowcol(cls, row: int, col: int, row_abs: bool, col_abs: bool):
        return cls(row, col, row_abs, col_abs)

    @property
    def cell(self) -> str:
        return xlcell.xl_rowcol_to_cell(
            self.row,
            self.col,
            self.row_abs,
            self.col_abs
        )

    @property
    def ref(self) -> str:
        return self.cell

    @property
    def col_name(self) -> str:
        return xlcell.xl_col_to_name(self.col, self.col_abs)

    def evolve(self, **kwargs) -> "XLCell":
        me = asdict(self)
        new = {
            **me,
            **{k: v for k, v in kwargs.items() if k in me}
        }
        return self.from_rowcol(**new)

    def __add__(self, other: "XLCell") -> "XLCell":
        return self.from_rowcol(
            self.row + other.row,
            self.col + other.col,
            self.row_abs,
            self.col_abs
        )

    def __sub__(self, other: "XLCell") -> "XLCell":
        return self.from_rowcol(
            self.row - other.row,
            self.col - other.col,
            self.row_abs,
            self.col_abs
        )

    def __call__(self) -> str:
        return self.cell


# %% XLRange
@dataclass(frozen=True)
class XLRange:
    cell1: XLCell
    cell2: XLCell

    @classmethod
    def from_rowcol(cls, row: int, col: int, row2:int, col2: int):
        return cls(
            XLCell(row, col),
            XLCell(row2, col2)
        )

    @classmethod
    def from_rowcol_abs(cls, row: int, col: int, row_abs: bool, col_abs: bool,
        row2: int, col2: int, row2_abs: bool, col2_abs: bool):
        return cls(
            XLCell(row, col, row_abs, col_abs),
            XLCell(row2, col2, row2_abs, col2_abs),
        )

    @classmethod
    def from_cells(cls, cell1: str, cell2: str):
        return cls(
            XLCell.from_cell(cell1),
            XLCell.from_cell(cell2)
        )

    @classmethod
    def from_ref(cls, ref: str):
        cell1, cell2 = ref.split(":")
        return cls.from_cells(cell1, cell2)

    @property
    def ref(self) -> str:
        return f"{self.cell1.cell}:{self.cell2.cell}"

    @property
    def columns(self) -> str:
        return f"{self.cell1.col_name}:{self.cell2.col_name}"

    def evolve(self, **kwargs):
        cell1 = asdict(self.cell1)
        cell2 = asdict(self.cell2)
        args1 = {
            k.replace("cell1_", ""): v
            for k, v in kwargs.items()
            if k.replace("cell1_", "") in cell1 and "cell1_" in k
        }
        args2 = {
            k.replace("cell2_", ""): v
            for k, v in kwargs.items()
            if k.replace("cell2_", "") in cell2 and "cell2_" in k
        }
        new1 = {**cell1, **args1}
        new2 = {**cell2, **args2}
        new2 = {
            k.replace("row", "row2").replace("col", "col2")
            : v for k, v in new2.items()
        }
        return self.from_rowcol_abs(**new1, **new2)

# %% xlformat
def xlformat(pattern: str, **kwargs) -> XLFormat:
    return XLFormat(pattern, kwargs)


# %% match_formats
def match_formats(data: pd.DataFrame, *formats: XLFormat) -> Dict[str,XLFormat]:
    matches = {}
    for col in data.columns:
        for fmt in formats:
            if col not in matches:
                pass
            if fnmatch(col, fmt.pattern):
                matches[col] = fmt
    return matches


# %% data_range
def data_range(data: pd.DataFrame, start: XLCell=XLCell(0, 0)) -> XLRange:
    """Returns the XLRange given the size of 'data' and 'start'."""
    rows, columns = data.shape

    # -1 from columns to make it 0-based, rows does not need this due to
    # having to account for header row
    end_cell = XLCell(rows, columns - 1)
    table_range = XLRange(start, start + end_cell)
    return table_range


# %%
def default_header_codes(dtype: str="object") -> Codes:
    align = (
        "left" if dtype == "object" else
        "left" if "date" in dtype else
        "right"
    )
    return {
        "align": align,
        "valign": "bottom",
        "bold": True,
        "text_wrap": True,
    }


def default_value_codes(dtype: str="object") -> Codes:
    num_format = (
        "" if dtype == "object" else
        "#,##0" if "int" in dtype else
        "#,##0.00" if "float" in dtype else
        "yyyy-mm-dd" if "date" in dtype else
        ""
    )
    align = (
        "left" if dtype == "object" else
        "left" if "date" in dtype else
        "right"
    )
    return {
        "align": align,
        "valign": "top",
        "num_format": num_format,
        "bold": False,
        "text_wrap": True,
    }


# %% column_formats
def column_formats(data: pd.DataFrame, *formats: XLFormat) -> Dict[str, ColFormat]:
    """Return a XLFormat object for each column in 'data'."""
    # Get matched formats
    matched = match_formats(data, *formats)

    result = {}
    # Get data type and set default format for all columns
    for header in data.columns:
        dtype = data[header].dtype.name.lower()
        value_fmt = default_value_codes(dtype)
        header_fmt = default_header_codes(dtype)

        # Override default formats with matched if they exist
        codes = matched[header].codes if header in matched else value_fmt
        result[header] = ColFormat(
            header_fmt,
            {**value_fmt, **codes},
        )
    return result


# %% to_xlsx
def to_xlsx(data: pd.DataFrame, file: pd.ExcelWriter, sheet_name: str,
    *formats: XLFormat, index: bool=False, start: XLCell=XLCell(0, 0),
    min_col_width: int=10, max_col_width: int=50,
    header_format: Optional[Codes]=None, make_table: bool=True,
    freeze_cell: XLCell=XLCell(1, 0)) -> XLRange:
    """Writes 'data' at 'start' and formats the columns."""

    # Get reference to sheet and workbook
    wb = file.book
    ws = wb.add_worksheet(sheet_name)
    table_range = data_range(data, start)

    # verify engine is xlsxwriter
    file.engine = "xlsxwriter"

    # Get column widths
    widths = calc_column_width(data, min_col_width, max_col_width)

    # Get formats
    fmts = column_formats(data, *formats)

    # Write data
    for idx, header in enumerate(data.columns):
        # Write Header
        header_format = (
            {**fmts[header].header_codes, **header_format}
            if header_format else
            fmts[header].header_codes
        )

        header_fmt = wb.add_format(header_format)
        ws.write_string(start.row, idx, header, header_fmt)

        # Write values
        value_fmt = wb.add_format(fmts[header].value_codes)
        ws.write_column(
            row=start.row + 1,
            col=idx,
            data=data[header],
            cell_format=value_fmt
        )

        # Set column width
        ws.set_column(idx, idx, widths[header])
        print(f"Wrote data for {header}")

    # Freeze panes
    freeze = start + freeze_cell
    ws.freeze_panes(freeze.row, freeze.col)
    print("Froze panes")

    # Add table
    if make_table:
        ws.add_table(
            table_range.ref,
            {
                "name": sheet_name.replace(" ", "_"),
                "header_row": True,
                "autofilter": False,
                "banded_rows": True,
                "style": "Table Style Light 1",
                "columns": [{"header": col} for col in data.columns],
            },
        )
        print("added table")

    return table_range



# %%
from pydantic.types import DirectoryPath
from sshtunnel import open_tunnel, SSHTunnelForwarder
from operator import methodcaller
import pymysql
import pandas as pd
import numpy as np
import requests
import toolz.curried as tz
from pathlib import Path
from typing import Dict, Any, Iterable, List, Callable, Type, Union
import sqlite3
from sqlite3 import Connection as SQLite
from pydantic import (
    BaseSettings,
    SecretStr,
    PositiveInt,
    FilePath,
    Field,
    DirectoryPath,
)
import logging
from atek.conversion import to_date, to_num
from atek.display import show

__all__ = [
    "Cache", "Domo", "domo", "MySQL", "mysql", "SQLite", "sqlite", "query",
]

# %%
# logging.basicConfig(
#     level=logging.INFO,
#     format="%(asctime)s %(message)s",
#     datefmt="%Y-%m-%d %H:%M:%S",
# )


# %%
class Cache(BaseSettings):
    """Returns the paths  using the ATEK_CACHE environment variable for the module."""
    cache: DirectoryPath
    report_dir: DirectoryPath

    class Config:
        env_prefix = "ATEK_"

    @property
    def query(self) -> DirectoryPath:
        return self.cache / "query"

    @property
    def secrets(self) -> DirectoryPath:
        return self.cache / "secrets"

    @property
    def mappings(self) -> DirectoryPath:
        return self.cache / "mappings"

    @property
    def report_dir(self) -> DirectoryPath:
        return self.report_dir


    def __truediv__(self, other: Union[Path,str]):
        return Path(self.cache) / other

    def __call__(self) -> DirectoryPath:
        return self.cache

# Cache()
# Cache() / "secrets"

# %%
class Domo(BaseSettings):
    """Connection object to a DOMO instance"""
    client_id: str
    secret: SecretStr
    dataset_id: str
    _header: Dict
    _url: str
    _data: List[Dict]
    _base_url: str="https://api.domo.com/v1/datasets"

    class Config:
        underscore_attrs_are_private = True
        env_prefix = "domo_"
        env_file = Cache().query
        secrets_dir = Cache().secrets

    def __init__(self, **data) -> None:
        super().__init__(**data)
        # Authenticate
        auth_url = (
            "https://api.domo.com/oauth/token?"
            "grant_type=client_credentials&scope=data"
        )
        auth = requests.auth.HTTPBasicAuth(
            self.client_id,
            self.secret.get_secret_value()
        )
        auth_response = requests.get(auth_url, auth=auth)
        token = auth_response.json()["access_token"]
        logging.info(f"AUTHENTICATED: Domo")
        logging.debug(self)

        # Assemble post url
        self._header = {"Authorization": f"bearer {token}"}
        self.connect(self.dataset_id)

    def connect(self, dataset_id: str) -> None:
        self._url = (
            f"{self._base_url}/query/execute/{dataset_id}"
            "?includeHeaders=true"
        )

    def execute(self, sql: str, *args, **kwargs) -> pd.DataFrame:
        """Stub method to be consistant with api for SQLite() and MySQL()
        classes."""
        logging.info(f"EXECUTING: Domo Query\n{sql}")
        results = requests.post(
            self._url,
            headers=self._header,
            json={"sql": sql}
            ).json()
        columns = results["columns"]
        rows = results["rows"]
        self._data = [
            dict(zip(columns, row))
            for row in rows
        ]
        logging.info(f"RETURNED: Domo Query")

    @property
    def connection(self):
        """Stub method to be consistant with api for SQLite() and MySQL()
        classes."""
        return self

    @property
    def conn(self):
        """Stub method to be consistant with api for SQLite() and MySQL()
        classes."""
        return self

    @property
    def description(self):
        """Stub method to allow class to be used in pandas.read_sql method."""
        if len(self._data) == 0:
            return tuple()
        return tuple(
            tuple([col, type(val), None, None, None, None, None])
            for col, val in self._data[0].items()
        )

    def cursor(self):
        """Stub method to allow class to be used in pandas.read_sql method."""
        return self

    def fetchall(self) -> List[Dict]:
        """Stub method to allow class to be used in pandas.read_sql method."""
        return [row for row in self._data]

    def close(self):
        """Stub method to allow class to be used in pandas.read_sql method."""
        pass

    def __iter__(self):
        for row in self._data:
            yield row

    def datasets(self) -> pd.DataFrame:
        """Retrieves a list of datasets from the domo instance for which the
        token has access."""
        limit = 50
        offset = 0

        # Cycle through the list in chunks of 'limit'
        rows = []
        while True:
            url = f"{self._base_url}?offset={offset}&limit={limit}"
            chunk = requests.get(url, headers=self._header).json()
            if not chunk: break
            for row in chunk:
                rows.append(row)
            offset += limit
        return pd.DataFrame.from_records(rows)

    def tables(self) -> List[str]:
        return self.datasets().name.tolist()

    def col_info(self, table_name: str) -> pd.DataFrame:
        dataset_id = tz.pipe(
            self.datasets()
            .query(f"name == '{table_name}' or id == '{table_name}'")
            .id.tolist()[0]
        )
        self.connect(dataset_id)
        return tz.pipe(
            self.query("select * from table limit 1")
            .dtypes
            .to_frame(name="data_type")
            .reset_index()
            .rename(columns={"index": "column"})
            .assign(**{
                "table": table_name,
                "dataset_id": dataset_id,
                "column_index": lambda df: np.arange(len(df)),
            })
            .filter(["dataset_id", "table", "column_index", "column", "data_type"])
        )

    def schema(self) -> pd.DataFrame:
        return tz.pipe(
            self.datasets()
            .id.tolist()
            ,tz.map(self.col_info)
            ,pd.concat
        )

    def query(self, sql: str, *args, **kwargs) -> pd.DataFrame:
        self.execute(sql)
        return tz.pipe(
            pd.DataFrame.from_records(self._data)
            ,to_date("Date")
            ,to_num("Fee|Amount")
        )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

def domo() -> Domo:
    return Domo()

def _dict_factory(row, cursor):
    """Returns a dict object for each row of a cursor with the keys = to the
    column names"""
    return {
        col[0]: row[idx]
        for idx, col in enumerate(cursor.description)
    }


# %%
class MySQL(BaseSettings):
    """Connection object to a mysql database accessed using a jump server and
    ssh keys"""
    remote_host: str="localhost"
    remote_port: PositiveInt=3306

    jump_host: str
    jump_port: PositiveInt=22
    jump_username: str
    jump_pkey: FilePath
    jump_password: SecretStr

    server_host: str
    server_port: PositiveInt=22
    server_username: str
    server_pkey: str
    server_password: SecretStr

    db_host: str="localhost"
    db_name: str
    db_username: str
    db_password: SecretStr

    kwargs: Dict=Field(default_factory=dict)

    _jump: SSHTunnelForwarder=None
    _server: SSHTunnelForwarder=None
    _connection: pymysql.connections.Connection=None
    _with_context: bool = False
    _cursor: pymysql.cursors.Cursor

    class Config:
        underscore_attrs_are_private = True
        env_prefix = "mysql_"
        env_file = Cache().query
        secrets_dir = Cache().secrets

    def __init__(self, **data) -> None:
        super().__init__(**data)

        self._jump = open_tunnel(
            ssh_address_or_host=(self.jump_host, self.jump_port),
            remote_bind_address=(self.remote_host, self.remote_port),
            ssh_pkey=str(self.jump_pkey),
            ssh_username=self.jump_username,
            ssh_password=self.jump_password.get_secret_value(),
        )
        self._jump.start()
        logging.info(f"OPENED: {self.jump_host}:{self.jump_port}")

        self._server = open_tunnel(
            ssh_address_or_host=(self.server_host, self.server_port),
            remote_bind_address=(self.remote_host, self.remote_port),
            ssh_pkey=self.server_pkey,
            ssh_username=self.server_username,
            ssh_password=self.server_password.get_secret_value(),
        )
        self._server.start()
        logging.info(f"OPENED: {self.server_host}:{self.server_port}")

        self._connection = pymysql.connect(
            host=self.db_host,
            db=self.db_name,
            user=self.db_username,
            password=self.db_password.get_secret_value(),
            port=self._server.local_bind_port,
            **self.kwargs
        )
        logging.info(f"OPENED: {self.db_host}:{self.db_name}")

    def close(self, *args, **kwargs) -> None:
        self._connection.close(*args, **kwargs)
        # self._connection = None
        logging.info(f"CLOSED: {self.db_host}:{self.db_name}")

        self._server.stop()
        self._server = None
        logging.info(
            f"CLOSED: {self.server_host}:{self.server_port}")

        self._jump.stop()
        self._jump = None
        logging.info(f"CLOSED: {self.jump_host}:{self.jump_port}")

    @property
    def connection(self) -> pymysql.connections.Connection:
        return self._connection

    @property
    def conn(self) -> pymysql.connections.Connection:
        return self.connection

    def __enter__(self):
        self._with_context = True
        return self

    def __exit__(self, *args):
        self.close()
        self._with_context = False

    def query(self, sql: str, *args, **kwargs) -> pd.DataFrame:
        logging.info(f"EXECUTING: {self.db_host}:{self.db_name} Query\n{sql}")
        df = pd.read_sql(sql, self.conn, *args, **kwargs)
        logging.info(f"RETURNED: {self.db_host}:{self.db_name} Query")
        if not self._with_context:
            self.close()
        return df

    def col_info(self, table_name: str) -> pd.DataFrame:
        return tz.pipe(
            self.query(f"describe {table_name}")
            .assign(**{
                "table": table_name,
                "column_index": lambda df: np.arange(len(df)),
            })
            .rename(columns=lambda x: (
                "column" if x == "Field" else
                "data_type" if x == "Type" else
                "nullable" if x == "Null" else
                x.lower()
            ))
            .filter(["table", "column_index", "column", "data_type", "nullable", "key", "default", "extra"])
        )

    def tables(self) -> List[str]:
        return tz.pipe(
            self.query("show tables")
            .rename(columns=lambda x: "table_name")
            .table_name.tolist()
        )

    def schema(self) -> pd.DataFrame:
        with self as con:
            return tz.pipe(
                con.tables()
                ,tz.map(con.col_info)
                ,pd.concat
            )

    def cursor(self) -> sqlite3.Cursor:
        self._cursor = self.conn.cursor()
        return self._cursor

    def execute(self, sql: str, *args, **kwargs):
        self.cursor()
        self._cursor.execute(sql, *args, **kwargs)

    def __iter__(self):
        for row in self._cursor:
            yield _dict_factory(row, self._cursor)

def mysql() -> MySQL:
    return MySQL()


# %%
class SQLite(BaseSettings):
    """Connection object to a sqlite database"""
    path: FilePath
    kwargs: Dict=Field(default_factory=dict)
    _connection: sqlite3.Connection=None
    _with_context: bool = False
    _cursor: sqlite3.Cursor

    class Config:
        underscore_attrs_are_private = True
        env_prefix = "sqlite_"
        env_file = Cache().query
        secrets_dir = Cache().secrets

    def __init__(self, **data) -> None:
        super().__init__(**data)
        self.connect()

    def connect(self) -> sqlite3.Connection:
        self._connection = sqlite3.connect(self.path, **self.kwargs)
        logging.info(f"OPENED: {self.path}")

    def close(self) -> None:
        self._connection.close()
        # self._connection = None
        logging.info(f"CLOSED: {self.path}")

    @property
    def connection(self) -> sqlite3.Connection:
        return self._connection

    @property
    def conn(self) -> sqlite3.Connection:
        return self.connection

    def __enter__(self):
        self._with_context = True
        return self

    def __exit__(self, *args):
        self.close()
        self._with_context = False

    def query(self, sql: str, *args, **kwargs) -> pd.DataFrame:
        logging.info(f"EXECUTING: {self.path} Query\n{sql}")
        df = pd.read_sql(sql, self.conn)
        logging.info(f"RETURNED: {self.path} Query")
        if not self._with_context:
            self.close()
        return df

    def tables(self) -> List[str]:
        return self.query("""
            select name as table_name
            from sqlite_master
            where type = 'table'
            """).table_name.tolist()

    def col_info(self, table_name: str) -> pd.DataFrame:
        if table_name not in self.tables():
            raise ValueError(f"{table_name} does not exist in {self.path}")

        self.connect()
        return tz.pipe(
            self.query(f"pragma table_info({table_name})")
            ,lambda df: df
            .assign(**{
                "table": table_name,
                "nullable": lambda df: np.where(df.notnull == 0, "YES", "NO")
            })
            .rename(columns={
                "name": "column",
                "type": "data_type",
                "cid": "column_index",
                "pk": "key",
                "dflt_value": "default",
            })
            .filter(["table", "column_index", "column", "data_type",
                "nullable", "default"])
        )

    def schema(self) -> pd.DataFrame:
        with self as con:
            return pd.concat([
                con.col_info(table)
                for table in con.tables()
            ])

    def cursor(self) -> sqlite3.Cursor:
        self.connect()
        self._cursor = self.conn.cursor()
        return self._cursor

    def execute(self, sql: str, *args, **kwargs):
        self.cursor()
        self._cursor.execute(sql, *args, **kwargs)

    def __iter__(self):
        for row in self._cursor:
            yield _dict_factory(row, self._cursor)


def sqlite() -> SQLite:
    return SQLite()

# %%
def query(connection: Union[MySQL,SQLite,Domo], sql: str, *args, **kwargs) -> List[Dict]:
    return connection.query(sql, *args, **kwargs)



from tabulate import tabulate
from atek.conversion import *
from typing import *
from textwrap import fill
from functools import singledispatch
import toolz.curried as tz
import pandas as pd

__all__ = ["show"]


def wrap(value: Any, width: int) -> str:
    return fill(str(value), width=width)


def wrapped(d: Dict[str, Any], width: int) -> Dict[str,str]:
    assert isinstance(d, dict), f"param 'd' must be a dict, provided {d}"
    return {
        wrap(k, width): wrap(v, width)
        for k, v in d.items()
    }


@singledispatch
def show(data: Iterable[Dict[str,Any]], *, caption: str="", limit: int=20,
    column_width: int=20, print_: bool=True, **tabulate_args
    ) -> str:

    # Default args for tabulate function
    default_args = {
        "tablefmt": "fancy_grid",
        "headers": "keys"
    }

    # Combine any passed args to default args for tabulate
    args = (
        tz.merge(default_args, tabulate_args) if tabulate_args else
        default_args
    )

    # Add caption if there is one
    table = f"\n{caption}\n" if caption else "\n"
    # Limit, wrap, and get results of tabulate
    table += tz.pipe(
        data
        ,tz.take(limit)
        ,tz.map(tz.partial(wrapped, width=column_width))
        ,list
        ,tz.partial(tabulate, **args)
    )

    # Returns the str result of tabulate if print_ set
    # Used for debugging and testing purposes
    if print_:
        print(table)
        return None
    return table


@show.register(dict)
@show.register(str)
@show.register(pd.DataFrame)
def _(data, **kwargs):
    return show(to_table(data), **kwargs)

show = tz.curry(show)


# %%
#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from fnmatch import fnmatch
from tabulate import tabulate
from textwrap import fill
from typing import *
import atek
from pathlib import Path
from dateutil.relativedelta import relativedelta as delta
from datetime import date, datetime
import pandas as pd
import re
import toolz.curried as tz
#  import atek
import censusdata
import us


__all__ = (
    "valid_path delta ymd bom eom notebook_setup states counties "
    "state_county period period_display"
).split()


# %%k
def valid_path(path: Union[str, Path]) -> Path:
    """Returns a valid pathlib.Path object or provides smart error message
    indicating which parts are valid (exist) and which parts do not exist."""
    _path = Path(path).expanduser()
    # If expanded path or directory exists then return the expanded path
    if _path.exists():
        return _path

    # If expanded path doesn't exist, return a helpful error message by
    # calculating the good part (exists) of the path and the bad part of
    # path
    temp = _path.parent
    while True:
        if temp.exists():
            bad_part = _path.relative_to(temp)
            raise ValueError(
                f"{_path} does not exist.\n"
                f"Good part = {temp}, \n"
                f"Bad part = {bad_part}"
            )
        temp = temp.parent


# %%
def notebook_setup():
    pd.set_option("display.max_columns", 300)
    pd.set_option("display.max_colwidth", 0)
    pd.set_option("display.max_rows", 1_000)
    pd.set_option("display.min_rows", 300)
    pd.set_option("display.date_yearfirst", True)
    pd.set_option("display.float_format", lambda value: f"{value:,.2f}")
    pd.set_option("display.precision", 2)


# %%
def ymd(dt: Optional[date]=None) -> str:
    """Formats a date as yyyy-mm-dd."""
    dt = dt or date.today()
    return dt.strftime("%Y-%m-%d")


# %%
def bom(dt: Optional[date]=None, months: int=0) -> date:
    """Returns the first day of the month.
    dt -> change day to 1 -> add months."""
    dt = dt or date.today()
    return dt + delta(day=1, months=months)


# %%
def eom(dt: Optional[date]=None, months: int=0) -> date:
    """Returns the last day of the month.
    dt -> change day to 1 -> add months + 1 -> subtract a day."""
    dt = dt or date.today()
    return dt + delta(day=1, months=months+1, days=-1)


# %%k
def states() -> pd.DataFrame:
    """Returns a pandas dataframe of State FIPS codes, abbreviations, and
    state names.

    Data comes from the acs5 2015 census data.
    """
    return tz.pipe(
        censusdata.geographies(
            censusdata.censusgeo([("state", "*")]),
            "acs5",
            2015
        )
        ,lambda d: [
            {
                "State": k,
                "State_FIPS": v.params()[0][1],
                "State_ID": int(v.params()[0][1]),
            }
            for k, v in d.items()
        ]
        ,tz.map(lambda d: tz.assoc(d, "State2", us.states.lookup(d["State"]).abbr))
        ,pd.DataFrame.from_records
    )


# %%
def counties() -> pd.DataFrame:
    """Returns a pandas dataframe of County Names, Cleaned County names.

    Cleaned County Names = Words Borough, Parish, City, County, Census Area,
        Municipality, and Municipio (PR) have been removed.

    Data comes from the acs5 2015 census data.
    """
    return tz.pipe(
        censusdata.geographies(
            censusdata.censusgeo([("state", "*"), ("county", "*")]),
            "acs5",
            2015
        )
        ,lambda d: [
            {
                "County_State": k,
                "State_FIPS": v.params()[0][1],
                "County_FIPS": v.params()[1][1],
                "County_ID": int(v.params()[1][1]),
            }
            for k, v in d.items()
        ]
        ,pd.DataFrame.from_records
        ,lambda df: df
        .assign(**{
            "County_Full": lambda df: df.County_State.str.extract("(^.*)(?=,.)"),
        })
        .assign(**{
            "County": lambda df: df.County_Full.str
                .replace(" County", "")
                .replace(" and Borough", "")
                .replace(" Census Area", "")
                .replace(" Municipality", "")
                .replace(" Borough", "")
                .replace(" Parish", "")
                .replace(" City", "")
                .replace(" Municipio", ""),
            "State_County_FIPS": lambda df: df.apply(lambda row:
                row["State_FIPS"] + row["County_FIPS"],
                axis=1
            ),
            "State_County_ID": lambda df: df.State_County_FIPS.astype("int"),
        })
        .drop(columns=["County_State"])
    )


# %k%
def state_county() -> pd.DataFrame:
    """Combines results of states() and counties() into 1 pandas dataframe.
    Data comes from the acs5 2015 census data.
    """
    return tz.pipe(
        states()
        .merge(counties(), on="State_FIPS", how="left")
        .assign(**{
            "State_County": lambda df: df.apply(lambda row:
                row["State2"] + "|" + row["County"],
                axis = 1,
            ),
        })
    )


# %%
@tz.curry
def period_display(min_fmt: str, max_fmt: str,
    min_date_col: pd.Series, max_date_col: pd.Series) -> pd.Series:
    return min_date_col.dt.strftime(min_fmt) + max_date_col.dt.strftime(max_fmt)


# %%
def period(period_format: str="%Y-%m", period_name: str="", *,
    start: str="2020-01", end: Optional[str]=None,
    period_to_date: bool=False, day_freq: str="D",
    make_table: bool=False, today: str="",
    period_display_func: Optional[Callable]=None,
    ) -> pd.DataFrame:
    """Creates a period dataset. Used to create the period table.str
    period_format is applied using the strftime method.
    display_func is a func that takes the period_min_date and peiod_max_date
        and returns a formatted string representing the period_display column.
    """

    # Create calendar day series and business series
    today = pd.Timestamp.today().date() if today == "" else pd.Timestamp(today)
    end = end or today + pd.DateOffset(day=1, years=1)
    days = pd.date_range(start, end, freq=day_freq)
    period_name = period_format.replace("%", "") if period_name == "" else period_name

    period_display_func = period_display_func or period_display("%m/%d-", "%m/%d")

    return tz.pipe(
        pd.DataFrame({
            "period_date": days,
            "days": 1,
            "today": today,
            "period_name": period_name,
        })

        # Calculate columns
        .assign(**{
            "period": lambda df: df.period_date.dt.strftime(period_format),
            "day3": lambda df: df.period_date.dt.strftime("%a"),
            # If process is run on a Sat and day_freq = "B", date for Today
            # will not be present to use next closest day befor actual today
            "is_today": lambda df: np.where(
                df.query("period_date <= today")
                .period_date.max() == df.period_date, 1, 0),

            # Period to date columns
            "days_of_period": lambda df: df.groupby("period").days
                .transform("cumsum"),
            "days_in_period": lambda df: df.groupby("period").days
                .transform("sum"),
            "remaining_days_in_period": lambda df:
                df.days_in_period - df.days_of_period,
            "period_to_date": lambda df: np.where(
                df.days_of_period <= df.query("is_today == 1")
                    .days_of_period.max(), 1, 0),
        })

        # period_to_date = True, filter by is_bus_days_to_date
        ,lambda df: (df.query("period_to_date == 1") if period_to_date else df)

        # Calculate min, max, start, and end dates for each period
        ,lambda df: df.assign(**{
            "period_min_date": lambda df: df.groupby("period")
                .period_date.transform("min"),
            "period_max_date": lambda df: df.groupby("period")
                .period_date.transform("max"),
            "period_display": lambda df:
                period_display_func(df.period_min_date, df.period_max_date),
                # df.period_min_date.dt.strftime("%m/%d-")
                # + df.period_max_date.dt.strftime("%m/%d"),
            "is_period_start": lambda df: np.where(
                df.period_min_date == df.period_date, 1, 0),
            "is_period_end": lambda df: np.where(
                df.period_max_date == df.period_date, 1, 0),

            # Relative period
            "row_id": lambda df: df.period.rank(method="dense"),
            "rel_period": lambda df: df.row_id - df.query("is_today == 1")
                .row_id.max(),

        })
        .drop(columns=["row_id", "days", "period_to_date"])
    )


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['atek']

package_data = \
{'': ['*']}

install_requires = \
['censusdata>=1.8,<2.0',
 'cytoolz>=0.10.1,<0.11.0',
 'jupyter>=1.0.0,<2.0.0',
 'openpyxl>=3.0.5,<4.0.0',
 'pandas>=1.1.0,<2.0.0',
 'pydantic[dotenv]>=1.7.2,<2.0.0',
 'pydomo>=0.2.3,<0.3.0',
 'pymysql>=0.10.0,<0.11.0',
 'requests>=2.24.0,<3.0.0',
 'sshtunnel>=0.1.5,<0.2.0',
 'tabulate>=0.8.7,<0.9.0',
 'toolz>=0.11.1,<0.12.0',
 'us>=2.0.2,<3.0.0',
 'xlrd>=1.2.0,<2.0.0',
 'xlsxwriter>=1.3.6,<2.0.0']

setup_kwargs = {
    'name': 'atek',
    'version': '0.1.42',
    'description': '',
    'long_description': None,
    'author': 'JediHero',
    'author_email': 'hansen.rusty@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Provides basic utilities to query a mysql database over an ssh tunnel"""
from sshtunnel import SSHTunnelForwarder, open_tunnel
import pandas as pd
import pymysql
from pymysql.cursors import Cursor, SSDictCursor
from pymysql.connections import Connection
from pathlib import Path
from typing import *
from configparser import ConfigParser
from contextlib import contextmanager
import os
import requests
from cytoolz.curried import pipe, partial, curry, get_in, take, map, merge

__all__ = "config query_atek02_main query_acad_domo".split()

def parse_config(config_path: str):
    """Reads a config file using configparser.ConfigParser from the standard library
    and converts any values with address in the key name to a ip address and port
    combination."""
    config = ConfigParser()
    path = Path(config_path).expanduser()
    config.read(path)
    settings = {
        section: {
            key: (
                tuple([
                    value.split()[0],
                    int(value.split()[1])
                ]) if "address" in key else
                value
            )
            for key, value in config[section].items()
        }
        for section in config.sections()
    }
    return settings


def config(env_var: str, *keys: str):
    """Returns the values in a config file as a dict or nested dict using an
    environment value ('env_var') which refers to a path.
    # 'path/to/some/config/file'
    [section 1]
    param1 = value1
    param2 = value2

    [section 2]
    parama = valuea
    paramb = valueb

    SOME_VAR='path/to/some/config/file'
    >>> config("SOME_VAR")
    {"section 1": {"param1": "value1", "param2": "value2"},
    "section 2": {"parama": "valuea", "paramb": "valueb"}}

    >>> config("SOME_VAR", "section 1")
    {"param1": "value1", "param2": "value2"}

    >>> config("SOME_VAR", "section 1", "param2")
    "value2"

    """
    # Get the enironment variable with a default of '' if it doesn't exist
    env_value = os.getenv(env_var, "")
    assert env_value != "", "f{env_var=} doesn't exist or returned ''"

    # Return the dict or potentially nested dict from parse_config
    settings = parse_config(env_value)

    # Return the value using nested keys
    return get_in([*keys], settings)


@contextmanager
def connect_atek02_main() -> Connection:
    """Creates a ssh tunnel and returns a connection options to atek02_main."""

    # Establish the ssh tunnel
    with SSHTunnelForwarder(**config("ATEK02_MAIN", "public_html")) as tunnel:

        # Create a connection to the database and a lazy cursor
        conn = pymysql.connect(
            **config("ATEK02_MAIN", "atek02_main"),
            port=tunnel.local_bind_port,
            cursorclass=SSDictCursor
        )

        yield conn

        conn.close()


Record = Dict[str,Any]
Table = Iterable[Record]

def query_atek02_main(sql: str) -> Table:
    """Returns a generator of dict objects as records."""

    with connect_atek02_main() as conn:
        cur = conn.cursor()
        cur.execute(sql)
        rows = cur.fetchall_unbuffered()
        for row in rows:
            yield row


def connect_acad_domo():
    """Returns an authorization header used in working with a domo
    instance api."""

    # Get client_id and token from config file and obtain an auth token
    domo = config("ACADEMY_DOMO", "acad_domo")
    auth = requests.auth.HTTPBasicAuth(domo["client_id"], domo["secret"])
    response = requests.get(domo["auth_url"], auth=auth)
    token = response.json()["access_token"]

    # Create and return the authorization header
    header = {"Authorization": f"bearer {token}"}
    return header


def datasets_acad_domo(name_or_id: Optional[str]=None,
    env_var: str="ACADEMY_DOMO") -> Table:
    """Retrieves a list of datasets from the domo instance for which the
    token has access."""

    header = connect_acad_domo()
    limit = 50
    offset = 0
    base_url = config(env_var, "acad_domo", "base_url")

    # Cycle through the list in chunks of 'limit'
    while True:
        url = f"{base_url}?offset={offset}&limit={limit}"
        chunk = requests.get(url, headers=header).json()
        if not chunk: break
        for row in chunk:
            # If name or id provided then only yield a row if it matches
            name = row["name"]
            id = row["id"]

            if name_or_id:
                if name_or_id == name or name_or_id == id:
                    yield row
                else:
                    pass

            # If name and id are not provided yield all rows
            else:
                yield row
        offset += limit


@curry
def query_acad_domo( name_or_id: Optional[str]=None, sql: Optional[str]=None):
    # Create the header
    header = connect_acad_domo()
    header["Accept"] = "application/json"
    name_or_id = name_or_id or "Prod - Appraisal - Mercury Appraisal No PDP"
    sql = sql or "select * from table"
    query = {"sql": sql}
    base_url = config("ACADEMY_DOMO", "acad_domo", "base_url")

    # Get thet table id
    table = list(datasets_acad_domo(name_or_id))
    assert len(table) > 0, "Returned dataset is empty... check the provided " \
        "'name_or_id' parameter."

    id = table[0]["id"]
    url = f"{base_url}/query/execute/{id}?includeHeaders=true"

    # Get the table in json formaatt
    response = requests.post(url, headers=header, json=query).json()

    # Assembe the table as a list of dict objects
    columns = response["columns"]
    rows = response["rows"]
    data = [
        dict(zip(columns, row))
        for row in rows
    ]

    # lazily return the data
    for row in data:
        yield row